import os

def create_technique_resource(material, resources):
	from .datafile import Resource, Statement
	tech_res = Resource(material.name+".tech", "technique")

	mat_res = resources[material.name+".mat"]

	st = Statement("pass", "")
	if mat_res:
		st.sub.append(tech_res.create_embed_statement("material", mat_res))

	if material.render_mode=='CUSTOM':
		shader = material.shader
		if shader.endswith(".glsl"):
			shader += ".shader"
		st.sub.append(Statement("shader", shader))

		if material.uniforms:
			ss = Statement("uniforms")
			for u in material.uniforms:
				ss.sub.append(Statement("uniform", u.name, *u.values[:u.size]))
			st.sub.append(ss)

	tech_res.statements.append(st)

	return tech_res

class MaterialExporter:
	def __init__(self):
		self.use_textures = True
		self.inline_texture_data = False

	def create_texture_exporter(self):
		from .export_texture import TextureExporter
		texture_export = TextureExporter()
		texture_export.inline_data = self.inline_texture_data
		return texture_export

	def export_technique_resources(self, material, resources):
		from .export_texture import SamplerExporter
		texture_export = self.create_texture_exporter()
		sampler_export = SamplerExporter()

		from .material import Material
		material = Material(material)

		if self.use_textures:
			for p in material.properties:
				if p.texture:
					tex_name = p.texture.image.name+".tex2d"
					if tex_name not in resources:
						resources[tex_name] = texture_export.export_texture(p.texture, p.tex_usage)

					samp_name = sampler_export.get_sampler_name(p.texture)
					if samp_name not in resources:
						resources[samp_name] = sampler_export.export_sampler(p.texture)

		mat_name = material.name+".mat"
		if mat_name not in resources:
			if material.type:
				resources[mat_name] = self.export_material(material, resources=resources)
			else:
				resources[mat_name] = None

	def export_technique(self, material, *, resources):
		return create_technique_resource(material, resources)

	def export_material(self, material, *, resources):
		from .datafile import Resource, Statement, Token
		mat_res = Resource(material.name+".mat", "material")

		if material.type!="pbr" and material.type!="unlit":
			raise Exception("Can't export unknown material type "+material.type)

		mat_res.statements.append(Statement("type", Token(material.type)));
		for p in material.properties:
			st = self.create_property_statement(mat_res, p, resources)
			if st:
				mat_res.statements.append(st)
		if self.use_textures:
			textures = [p.texture for p in material.properties if p.texture]
			if textures and not textures[0].default_filter:
				from .export_texture import SamplerExporter
				sampler_export = SamplerExporter()
				mat_res.statements.append(mat_res.create_reference_statement("sampler", resources[sampler_export.get_sampler_name(textures[0])]))

		return mat_res

	def create_property_statement(self, mat_res, prop, resources):
		from .datafile import Statement
		if self.use_textures and prop.texture:
			tex_res = resources[prop.texture.image.name+".tex2d"]
			from .util import basename
			fn = basename(prop.texture.image.filepath)
			if prop.texture.default_filter and fn:
				return Statement(prop.tex_keyword, fn)
			else:
				return mat_res.create_reference_statement(prop.tex_keyword, tex_res)
		elif not prop.keyword:
			return
		elif type(prop.value)==tuple:
			return Statement(prop.keyword, *prop.value)
		else:
			return Statement(prop.keyword, prop.value)


class MaterialAtlasExporter:
	def __init__(self):
		pass

	def export_technique_resources(self, material_atlas, resources):
		from .datafile import Resource, Statement, Token
		base_color_name = material_atlas.name+"_base_color.tex2d"
		base_color_res = resources.get(base_color_name)
		if not base_color_res:
			base_color_res = Resource(base_color_name, "texture2d")

			base_color_res.statements.append(Statement("storage", Token('SRGB_ALPHA'), *material_atlas.size))
			base_color_res.statements.append(Statement("raw_data", material_atlas.base_color_data))

			resources[base_color_name] = base_color_res

		sampler_name = "nearest.samp"
		sampler_res = resources.get(sampler_name)
		if not sampler_res:
			sampler_res = Resource(sampler_name, "sampler")

			sampler_res.statements.append(Statement("filter", Token('NEAREST')))

			resources[sampler_name] = sampler_res

		mat_name = material_atlas.name+".mat"
		if mat_name not in resources:
			mat_res = Resource(mat_name, "material")
			mat_res.statements.append(Statement("type", Token('pbr')))
			mat_res.statements.append(mat_res.create_reference_statement("base_color_map", base_color_res))
			mat_res.statements.append(mat_res.create_reference_statement("sampler", sampler_res))

			resources[mat_name] = mat_res

	def export_technique(self, material_atlas, *, resources):
		return create_technique_resource(material_atlas, resources)
