import math
import os

class SceneExporter:
	def __init__(self):
		self.selected_only = False
		self.visible_collections = True
		self.resource_collection = True
		self.skip_existing = True
		self.show_progress = True

	def export_to_file(self, context, out_fn):
		if self.selected_only:
			objs = context.selected_objects
		else:
			objs = context.scene.objects
		if self.visible_collections:
			collections = [c.collection for c in context.view_layer.layer_collection.children if not (c.hide_viewport or c.collection.hide_viewport)]
			objs = [o for o in objs if any((o.name in c.all_objects) for c in collections)]
		objs = [o for o in objs if o.type=="MESH" and not o.lod_for_parent]
		objs = [o for o in objs if (not o.compound or o.parent not in objs)]
		objs.sort(key=lambda x:x.name)

		path, base = os.path.split(out_fn)
		base, ext = os.path.splitext(base)

		from .export_object import ObjectExporter
		object_export = ObjectExporter()

		object_prototypes = {}
		unique_objects = []
		export_names = {}
		for o in objs:
			if o.name in object_prototypes:
				continue

			clones = [o]
			if not any(s.link=="OBJECT" for s in o.material_slots):
				for u in objs:
					if u is o:
						continue
					if u.data.name!=o.data.name:
						continue
					if any(s.link=="OBJECT" for s in u.material_slots):
						continue

					clones.append(u)

			prefix = o.name
			for c in clones:
				while not c.name.startswith(prefix):
					pos = max(prefix.rfind(' '), prefix.rfind('.'))
					if pos<0:
						break;
					prefix = prefix[:pos]

			if prefix:
				export_names[o.name+".object"] = prefix.strip(" .")+".object"

			unique_objects.append(o)
			for c in clones:
				object_prototypes[c.name] = o

		from .util import Progress
		progress = Progress(self.show_progress and context)

		resources = {}
		self.export_scene_resources(context, unique_objects, resources, progress)
		for n, r in resources.items():
			if r.name in export_names:
				r.name = export_names[r.name]

		scene_res = self.export_scene(context, objs, progress, prototypes=object_prototypes, resources=resources)
		refs = scene_res.collect_references()

		if self.resource_collection:
			filter = None
			if self.skip_existing:
				filter = lambda r: not os.path.exists(os.path.join(path, r.name))
			scene_res.write_collection(os.path.join(path, base+"_resources.mdc"), exclude_self=True, filter=filter)
		else:
			res_dir = os.path.join(path, base+"_resources")
			if not os.path.exists(res_dir):
				os.makedirs(res_dir)
			for r in refs:
				r.write_to_file(os.path.join(res_dir, r.name))

		scene_res.write_to_file(out_fn)

	def export_scene_resources(self, context, objs, resources, progress):
		from .export_object import ObjectExporter
		object_export = ObjectExporter()
		object_export.single_file = False

		material_atlases = {}

		for i, o in enumerate(objs):
			progress.push_task_slice(o.name, i, len(objs))
			object_export.export_object_resources(context, o, resources, progress, material_atlases=material_atlases)
			obj_name = o.name+".object"
			resources[obj_name] = object_export.export_object(context, o, progress, resources=resources)
			progress.pop_task()

	def export_scene(self, context, objs, progress, *, prototypes, resources):
		from .datafile import Resource, Statement
		scene_res = Resource("scene.scene", "scene")

		for o in objs:
			obj_res = resources[prototypes[o.name].name+".object"]
			st = scene_res.create_reference_statement("object", obj_res, o.name)

			ss = Statement("transform")

			loc = o.matrix_world.to_translation()
			ss.sub.append(Statement("position", *tuple(loc)))

			quat = o.matrix_world.to_quaternion()
			if o.rotation_mode in ('XYZ', 'XZY', 'YXZ', 'YZX', 'ZXY', 'ZYX'):
				angles = [a*180/math.pi for a in quat.to_euler()]
				ss.sub.append(Statement("euler", *angles));
			else:
				ss.sub.append(Statement("rotation", quat.angle*180/math.pi, *tuple(quat.axis)))

			scale = o.matrix_world.to_scale()
			ss.sub.append(Statement("scale", *tuple(scale)))

			st.sub.append(ss)
			scene_res.statements.append(st)

		progress.set_progress(1.0)

		return scene_res
