#include <msp/core/application.h>
#include <msp/core/getopt.h>
#include <msp/gl/glsl/compiler.h>
#include <msp/gl/glsl/glsl_error.h>
#include <msp/io/print.h>
#include <msp/strings/utils.h>

class GlslCompiler: public Msp::RegisteredApplication<GlslCompiler>
{
private:
	std::string source_fn;
	Msp::GL::SL::Features features;
	Msp::GL::SL::Compiler::Mode compile_mode;
	std::map<std::string, int> spec_values;
	bool parse_only;
	bool combined;
	Msp::GL::SL::Stage::Type stage;
	bool dump_ast;
	std::string out_filename;

public:
	GlslCompiler(int, char **);

	virtual int main();
};

using namespace std;
using namespace Msp;

GlslCompiler::GlslCompiler(int argc, char **argv):
	features(GL::SL::Features::latest()),
	compile_mode(GL::SL::Compiler::PROGRAM),
	parse_only(false),
	combined(false),
	stage(GL::SL::Stage::SHARED),
	dump_ast(false)
{
	string stage_str;
	vector<string> spec_values_in;
	unsigned as_module = 0;
	string module_type = "glsl";
	unsigned target_version = 0;

	GetOpt getopt;
	getopt.add_option('c', "combined", combined, GetOpt::NO_ARG).set_help("Output combined GLSL");
	getopt.add_option('a', "dump-ast", dump_ast, GetOpt::NO_ARG).set_help("Dump AST for debugging");
	getopt.add_option('p', "parse_only", parse_only, GetOpt::NO_ARG).set_help("Only parse the loaded source (implies -a)");
	getopt.add_option('e', "specialize", spec_values_in, GetOpt::REQUIRED_ARG).set_help("Set specialization constant", "NAME:VALUE");
	getopt.add_option('s', "stage", stage_str, GetOpt::REQUIRED_ARG).set_help("Output GLSL for STAGE", "STAGE");
	getopt.add_option('m', "module", module_type, GetOpt::OPTIONAL_ARG).bind_seen_count(as_module).set_help("Compile as unspecialized module");
	getopt.add_option('t', "target-version", target_version, GetOpt::REQUIRED_ARG).set_help("Specify target GLSL version", "VER");
	getopt.add_option('o', "out-file", out_filename, GetOpt::REQUIRED_ARG).set_help("Write output to file instead of stdout", "FILE");
	getopt.add_argument("source", source_fn, GetOpt::REQUIRED_ARG).set_help("GLSL file to compile");
	getopt(argc, argv);

	if(target_version)
		features = GL::SL::Features::from_version(GL::Version(target_version/100, target_version%100));

	if(as_module)
	{
		if(module_type=="glsl" || module_type=="GLSL")
			compile_mode = GL::SL::Compiler::MODULE;
		else if(module_type=="spirv" || module_type=="spir-v" || module_type=="SPIRV" || module_type=="SPIR-V")
			compile_mode = GL::SL::Compiler::SPIRV;
		else
			throw usage_error("Invalid module type");
	}

	if(compile_mode==GL::SL::Compiler::SPIRV && out_filename.empty())
		throw usage_error("-o is required for SPIR-V");

	if(parse_only)
	{
		if(!stage_str.empty())
			throw usage_error("-s can't be used with -p");
		dump_ast = true;
	}

	if(!stage_str.empty() && compile_mode==GL::SL::Compiler::SPIRV)
		throw usage_error("-s can't be used with SPIR-V");
	else if(stage_str=="vertex")
		stage = GL::SL::Stage::VERTEX;
	else if(stage_str=="geometry")
		stage = GL::SL::Stage::GEOMETRY;
	else if(stage_str=="fragment")
		stage = GL::SL::Stage::FRAGMENT;
	else if(!dump_ast)
		combined = true;

	if(!spec_values_in.empty() && as_module)
		throw usage_error("Modules can't be specialized");
	for(vector<string>::const_iterator i=spec_values_in.begin(); i!=spec_values_in.end(); ++i)
	{
		unsigned colon = i->find(':');
		if(colon==string::npos || colon==0 || colon+1>=i->size())
			throw usage_error("Invalid specialization value");

		string value_str = i->substr(colon+1);
		int value;
		if(isnumrc(value_str))
			value = lexical_cast<int>(value_str);
		else
			value = lexical_cast<bool>(value_str);
		spec_values[i->substr(0, colon)] = value;
	}
}

int GlslCompiler::main()
{
	GL::SL::Compiler compiler(features);
	IO::File file(source_fn);
	compiler.load_source(file, source_fn);
	if(compile_mode==GL::SL::Compiler::PROGRAM)
		compiler.specialize(spec_values);
	if(!parse_only)
	{
		try
		{
			compiler.compile(compile_mode);
			string diag = compiler.get_diagnostics();
			if(!diag.empty())
				IO::print("Diagnostic messages from compiler:\n%s\n", diag);
		}
		catch(const GL::SL::invalid_shader_source &exc)
		{
			if(!dump_ast)
				throw;

			IO::print("Compilation resulted in errors:\n%s\n", exc.what());
			combined = false;
			stage = GL::SL::Stage::SHARED;
		}
	}

	if(dump_ast)
	{
		vector<GL::SL::Stage::Type> stages = compiler.get_stages();
		for(vector<GL::SL::Stage::Type>::const_iterator i=stages.begin(); i!=stages.end(); ++i)
			IO::print("%s\n", compiler.get_stage_debug(*i));
	}

	IO::Base *out = &IO::cout;
	RefPtr<IO::File> out_file;
	if(!out_filename.empty())
	{
		out_file = new IO::File(out_filename, IO::M_WRITE);
		out = out_file.get();
	}

	if(compile_mode==GL::SL::Compiler::SPIRV)
	{
		vector<UInt32> code = compiler.get_combined_spirv();
		out->write(reinterpret_cast<char *>(&code.front()), code.size()*4);
	}
	else if(combined)
		IO::print(*out, "%s\n", compiler.get_combined_glsl());
	else if(stage!=GL::SL::Stage::SHARED)
		IO::print(*out, "%s\n", compiler.get_stage_glsl(stage));

	return 0;
}
